/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

#include "bsp.h"
#include "gps.h"
#include "string.h"


static uint8_t END_OF_MESSAGE = '\0';
static uint8_t DIVIDER = ',';

// The parsed data
static gpsData data;


/*****************************************************************************
 ** Function name:               uartReadByteBlocking
 **
 ** Descriptions:                Blocking read of a byte from the GPS's UART.
 **
 ** parameters:                  None
 ** Returned value:              The read byte
 **
 *****************************************************************************/
static uint8_t uartReadByteBlocking(void)
{
	uint8_t ch = 0;
	while(bsp_rf_receive(&ch, 1) != 1) {
	}

	return ch;
}

/*****************************************************************************
 ** Function name:               hasPattern
 **
 ** Descriptions:                Tests if pBuf starts with pPattern.
 **
 ** parameters:                  Buffer to search and pattern to match
 ** Returned value:              1 if pBuf starts with pPattern, 0 otherwise
 **
 *****************************************************************************/
static uint8_t hasPattern(uint8_t *pBuf, uint8_t *pPattern)
{
	while(*pBuf != END_OF_MESSAGE && *pPattern != END_OF_MESSAGE){
		if(*pBuf != *pPattern){
			return 0;
		}
		pPattern++;
		pBuf++;
	}

	return 1;
}

/*****************************************************************************
 ** Function name:               pointToNextValue
 **
 ** Descriptions:                Moves past the next divider
 **
 ** parameters:                  Pointer to the string to search
 ** Returned value:              None
 **
 *****************************************************************************/
static void pointToNextValue(uint8_t **ppBuf)
{
	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == DIVIDER) {
			(*ppBuf)++; // point to the start of next value
			break;
		}
		(*ppBuf)++;
	}
}

/*****************************************************************************
 ** Function name:               convertCordinateToDegree
 **
 ** Descriptions:                Converts the pBuf string which is in the
 **                              "ddmm.mmmm" format into an integer representation
 **
 ** parameters:                  The buffer, the resulting integer and the
 **                              length of the buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void convertCordinateToDegree(uint8_t *pBuf, int* pDegree, int len)
{
	int index = 0;
	int sum = 0;
	int deg = 0;
	int min = 0;
	int div = 0;
	int pow = 1;

	 for (index = len; index >=0; index--) {
		if (pBuf[index] == '.') {
			div = 1;
			continue;
		}
		sum += pow * (pBuf[index] & 0x0F);

		if (index > 0) {
			pow *= 10;
			div *= 10;
		}
	}

	div = pow / div;
	deg = sum / (div*100);
	min = sum - (deg*div*100);
	// convert to decimal minutes
	min = (min * 100) / 60;
	*pDegree = (deg*div*100) + min;

	if (div > 10000) {
		 // normalize minutes to 6 decimal places
		*pDegree /= (div / 10000);
	}

}

/*****************************************************************************
 ** Function name:               parseUTC
 **
 ** Descriptions:                Extracts the UTC time string in hhmmss.sss,
 **                              ignoring the .sss part and stores the result
 **                              as a string in data.utcTime.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseUTC(uint8_t **ppBuf)
{
	int index = 0;

	// parse utc hhmmss.sss
	while(**ppBuf != END_OF_MESSAGE) {
		if(**ppBuf == '.') {
			pointToNextValue(ppBuf);
			break; //reached end of the value
		}

		data.utcTime[index++] = **ppBuf;

		if(index == 2 || index == 5) {
			//Add divider
			data.utcTime[index++] = ':';
		}
		(*ppBuf)++;
	}
	data.utcTime[index] = '\0';
}

/*****************************************************************************
 ** Function name:               parseLatitude
 **
 ** Descriptions:                Extracts the latitude information and stores
 **                              the result as an integer in data.latitude.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseLatitude(uint8_t **ppBuf)
{
	int index = 0;

	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == DIVIDER) {
			(*ppBuf)++;  //reached end of the value
			break;
		}
		data.bufLatitude[index++] = **ppBuf;
		(*ppBuf)++;
	}

	convertCordinateToDegree((uint8_t *) &data.bufLatitude, &data.latitude, 8);
}

/*****************************************************************************
 ** Function name:               parseLongitude
 **
 ** Descriptions:                Extracts the longitude information and stores
 **                              the result as an integer in data.longitude.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseLongitude(uint8_t **ppBuf)
{
	int index = 0;

	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == DIVIDER) {
			(*ppBuf)++;  //reached end of the value
			break;
		}
		data.bufLongitude[index++] = **ppBuf;
		(*ppBuf)++;
	}

	convertCordinateToDegree((uint8_t *) &data.bufLongitude, &data.longitude, 9);
}

/*****************************************************************************
 ** Function name:               parseNSIndicator
 **
 ** Descriptions:                Extracts the North/South information and
 **                              modifies the data.latitude accordingly.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseNSIndicator(uint8_t **ppBuf)
{
	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == 'S') {
			data.latitude = -data.latitude;
			pointToNextValue(ppBuf);
			break;
		}
		else if (**ppBuf == DIVIDER) {
			(*ppBuf)++; //reached end of the value
			break;
		}
		(*ppBuf)++;
	}
}

/*****************************************************************************
 ** Function name:               parseEWIndicator
 **
 ** Descriptions:                Extracts the East/West information and
 **                              modifies the data.longitude accordingly.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseEWIndicator(uint8_t **ppBuf)
{
	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == 'W') {
			data.longitude = -data.longitude;
			pointToNextValue(ppBuf);
			break;
		}
		else if (**ppBuf == DIVIDER) {
			(*ppBuf)++; //reached end of the value
			break;
		}
		(*ppBuf)++;
	}
}

/*****************************************************************************
 ** Function name:               parseFixIndicator
 **
 ** Descriptions:                Extracts the position fix indicator and
 **                              stores it as an integer in data.positionFixed.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseFixIndicator(uint8_t **ppBuf)
{
	data.positionFixed = (**ppBuf) & 0x0F;
	pointToNextValue(ppBuf);
}

/*****************************************************************************
 ** Function name:               parseSatellites
 **
 ** Descriptions:                Extracts the number of satellites and
 **                              stores it as a string in data.satellitesUsed.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseSatellites(uint8_t **ppBuf)
{
	int index = 0;

	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == DIVIDER) {
			(*ppBuf)++;  //reached end of the value
			break;
		}
		data.satellitesUsed[index++] = **ppBuf;
		(*ppBuf)++;
	}
	data.satellitesUsed[index] = '\0';
}

/*****************************************************************************
 ** Function name:               parseHDOP
 **
 ** Descriptions:                Extracts the horizontal dilution of altitude.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseHDOP(uint8_t **ppBuf)
{
	//We just skip this
	pointToNextValue(ppBuf);
}

/*****************************************************************************
 ** Function name:               parseAltitudes
 **
 ** Descriptions:                Extracts the altitude information and
 **                              stores it as a string in data.altitude.
 **
 ** parameters:                  The buffer
 ** Returned value:              None
 **
 *****************************************************************************/
static void parseAltitudes(uint8_t **ppBuf)
{
	int index = 0;
	// parse altitude
	while(**ppBuf != END_OF_MESSAGE) {
		if (**ppBuf == DIVIDER) {
			(*ppBuf)++;  //reached end of the value
			break;
		}
		data.altitude[index++] = **ppBuf;
		(*ppBuf)++;
	}
	data.altitude[index] = '\0';
}

/*****************************************************************************
 ** Function name:               GPSResetData
 **
 ** Descriptions:                Clears the gathered data
 **
 ** parameters:                  None
 ** Returned value:              None
 **
 *****************************************************************************/
void GPSResetData(void)
{
	memset(&data, 0, sizeof(gpsData));
}

/*****************************************************************************
 ** Function name:               GPSRetreiveData
 **
 ** Descriptions:                Reads and parses the next set of GPS data.
 **
 ** parameters:                  None
 ** Returned value:              The parsed information
 **
 *****************************************************************************/
const gpsData* GPSRetreiveData(void)
{
	uint8_t * pattern = (uint8_t*)"GPGGA";

	while (1) {
		uint8_t buf[100];
		uint8_t ch = 0;
		uint8_t *ptr = 0;
		int index = 0;

		// Retrieve the first byte
		if (bsp_rf_receive(&ch, 1) == 0)
			continue;

		 // look for "$GPGGA," header
		if (ch != '$') {
			continue;
		}

		// Retrieve the next six bytes
		for (index=0; index<6; index++) {
			buf[index] = uartReadByteBlocking();
		}

		//Check if its Global Positioning System fixed Data
		if(hasPattern((uint8_t*)&buf, pattern) == 0) {
			continue;
		}

		//Retrieve the data from the GPS module
		for (index=0; index<100; index++) {
			buf[index] = uartReadByteBlocking();

			if (buf[index] == '\r') {
				buf[index] = END_OF_MESSAGE;
				break;
			}
		}

		ptr = &buf[0];

		//parse UTC time
		parseUTC(&ptr);

		//parse Latitude
		parseLatitude(&ptr);

		// parse N/S field
		parseNSIndicator(&ptr);

		// parse Longitude
		parseLongitude(&ptr);

		// parse E/W field
		parseEWIndicator(&ptr);

		// parse fix position
		parseFixIndicator(&ptr);

		// parse satellites
		parseSatellites(&ptr);

		// parse horizontal-dilution-of-precision field
		parseHDOP(&ptr);

		// parse altitude
		parseAltitudes(&ptr);

		break;
	}
	return &data;
}
